open Absyn 
open Env
open EoplPrint
open List
open Format

exception TypeError of string
exception TODO

(* Type environment *)
type tenv = typ Env.env

(* A constraint set *) 
type constr = (typ * typ) list
(* A substitution list *)
type subst  = (var * typ) list

(* print out a constraint set, for debugging. *)
let print_constr (c:constr) :unit =
  print_list c 
	 (fun (t1,t2) -> 
		 print_string (string_of_typ t1);
		 print_string "=";
		 print_string (string_of_typ t2))
	 print_comma

(* print out a substitution, also for debugging. *)
let print_subst (s:subst) : unit = 
  print_list s
	 (fun (x,t) -> 
		 print_string (x ^ "->");
		 print_string (string_of_typ t))
	 print_comma

(* Create a new type (variable) *)
let newvar = 
  let ctr = ref 0 in 
	 fun () -> 
		ctr := !ctr + 1;
		VarType ("ty$"^ (string_of_int !ctr))
  

(* Determine the free variables of a type expression *)
let rec free_vars (t:typ) : var list = raise TODO
				 
(* Apply the substitution to the type expression *)
let rec subst_typ (s:subst) (t:typ) :typ =  raise TODO

(* convert a substitution into a constraint set *)
let subst2constr (s:subst) : constr = 
  map (fun (x,t) -> (VarType x,t)) s

(* Create a substitution from a constraint set. 
   The expression e is used to create a better error 
   message. *)
let rec unify (e:exp) (c:constr) : subst = raise TODO

  
(* Given a typing annotation, if there is no type listed, create a new 
 *  type variable and set that annotation to be equal to that variable. 
 *) 
let get_typ (tyopt:annot) : typ = 
  match !tyopt with 
		Some typ -> typ
	 | None -> let t = newvar () in (tyopt := Some t; t)


(* Type check a primitive, given a list of arguments. *)
let rec check_prim (p:prim) (args:typ list)  : typ * constr = 
  match p,args with 
      Plus,   [t1;t2] -> IntType, [(t1,IntType);(t2,IntType)]
    | Times,  [t1;t2] -> IntType, [(t1,IntType);(t2,IntType)]
    | Minus,  [t1;t2] -> IntType, [(t1,IntType);(t2,IntType)]
    | And,    [t1;t2] -> BoolType,[(t1,BoolType);(t2,BoolType)]
    | Or,     [t1;t2] -> BoolType, [(t1,BoolType);(t2,BoolType)]
    | Not,    [t] -> BoolType, [t,BoolType]
    | Add1,   [t] -> IntType, [t,IntType]
    | Sub1,   [t] -> IntType, [t,IntType]
    | Zero,   [t] -> BoolType, [t,IntType]
    | LT,     [t1;t2] -> BoolType,[(t1,IntType);(t2,IntType) ]     
    | GT,     [t1;t2] -> BoolType, [(t1,IntType);(t2,IntType)]
    | Equal,  [t1;t2] -> BoolType,[(t1,IntType);(t2,IntType)]

    | Pair,   [t1;t2] -> raise TODO
    | Fst,    [t] -> raise TODO
    | Snd,    [t] -> raise TODO
    | Hd,     [t] -> raise TODO
    | Tl,     [t] -> raise TODO
    | Cons,   [t1; t] -> raise TODO 
    | IsNull, [t] -> raise TODO

    | _, _ -> raise (TypeError "Unimplemented  primitive or invalid number of arguments")


(* Extend the type environment for a letrec expression and check that 
	all of the rhs's are well-typed.
	Returns the new environment and a constraint set for the expression.
*)
let rec extend_env_letrec 
  (tys:annot list)
  (proc_names:var list)
  (arg_typs: annot list list) 
  (idss: var list list) 
  (bodies: exp list)
  (env: tenv) : tenv * constr = raise TODO


and type_exp_aux (exp:exp) (env:tenv) : typ * constr = 
  match exp with 
      LitInt i -> IntType, []

    | LitBool b -> BoolType, []

    | LitNull ->  raise TODO

    | Var v ->  raise TODO

    | Prim (p, rands) -> 
        let args,c1 = type_exp_rands rands env in
        let ret,c2 = check_prim p args in 
			 ret, rev_append c2 c1

    | If (test_exp, true_exp, false_exp) -> raise TODO

	 | Let (vars,rands,body) -> 
        let arg_typs,arg_constr = type_exp_rands rands env in 
		  let body_typ,body_constr = type_exp body (extend_env vars arg_typs env) in  
			 body_typ, rev_append arg_constr body_constr

    | Lambda (tyopts,ids,body) -> raise TODO

    | App (rator,rands) -> raise TODO

    | Letrec (tys, proc_names, arg_tys, idss, bodies, letrec_body) -> 
		  let new_env,constr1 = extend_env_letrec tys proc_names arg_tys idss bodies env in
		  let body_typ,constr2 = type_exp letrec_body new_env in 
			 body_typ, (append constr1 constr2)

    | Varassign (id, rhs) -> raise TODO

    | Begin(exp, exps) -> 
        let v,constr = type_exp exp env in
        let rec loop l = 
          match l with 
              [] -> v,[]
            | [e] -> type_exp e env 
				| (hd :: tl) -> 
					 let _, constr1 = type_exp hd env in
					 let typ, constr2 = loop tl in 
						typ, rev_append constr1 constr2 in
          loop exps
            
and type_exp_rands (rands:exp list) (env:tenv): typ list * constr = 
  match rands with 
		[] -> [], []
	 | (rand::rands) -> 
		  let rt,ct = type_exp rand env in 
		  let rts,cts = type_exp_rands rands env in 
			 (rt::rts, rev_append ct cts)

and type_exp (exp:exp) (env:tenv) : typ * constr = 
  let typ,constr = type_exp_aux exp env in 
  let subst = unify exp constr in 
	 subst_typ subst typ, subst2constr subst


(* Apply a substitution to a typing annotation, 
 * setting that annotation to be the result of the 
 *	substitution *)
let subst_annot (s:subst) (tr:annot) : unit = 
  match (!tr) with 
		Some t -> (tr := Some (subst_typ s t))
	 | None -> failwith "Should be types everywhere."

(* Apply a substitution to an expression, updating 
 * all of the typing annotations. *)
let rec subst_type_exp (s:subst) (exp:exp) : unit = 
  match exp with 
      LitInt i -> ()
    | LitBool b -> ()
    | LitNull ->  ()
    | Var v -> ()
    | Prim (p, rands) -> iter (subst_type_exp s) rands
    | If (test_exp, true_exp, false_exp) -> 
		  begin
			 subst_type_exp s test_exp;
			 subst_type_exp s true_exp;
			 subst_type_exp s false_exp;
		  end
	 | Let (vars,rands,body) -> 
        (iter (subst_type_exp s) rands;
			subst_type_exp s body)
    | Lambda (tyopts,ids,body) ->    
		  (iter (subst_annot s) tyopts; 
			subst_type_exp s body)
    | App (rator,rands) -> 
		  (subst_type_exp s rator;
			iter (subst_type_exp s) rands)		  
    | Letrec (tys, proc_names, arg_tys, idss, bodies, letrec_body) -> 
		  iter (subst_annot s) tys;
		  iter (iter (subst_annot s)) arg_tys;
		  iter (subst_type_exp s) bodies;
		  subst_type_exp s letrec_body
    | Varassign (id, rhs) -> 
		 subst_type_exp s rhs 
    | Begin(exp, exps) -> 
		  (subst_type_exp s exp;
		  iter (subst_type_exp s) exps)

let type_program (e:exp) : typ = 
  let typ,constr = type_exp e Empty in
  let subst = unify e constr in 
 	 subst_type_exp subst e; 
	 subst_typ subst typ 
